/* YaRock Player
* Copyright (C) 2010-2013 Sebastien amardeilh (sebastien.amardeilh+yarock@gmail.com)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "AlbumCover/coverthread.h"

#include "AlbumCover/coverprovider_lastfm.h"
#include "AlbumCover/coverprovider_discogs.h"

#include "Database/database.h"
#include "MediaItem/mediaitem.h"
#include "utilities.h"
#include "debug.h"

#include <QSqlField>
#include <QSqlQuery>
#include <QSqlDriver>
#include <QVariant>

#include <QFile>
#include <QImage>


/*
********************************************************************************
*                                                                              *
*    Class CoverThread                                                         *
*                                                                              *
********************************************************************************
*/   

CoverThread::CoverThread(QObject *parent) : QThread(parent) 
{
    setObjectName("coverthread");
    setTerminationEnabled(true);

    //! init 
    m_current_request_id  = 0;
    m_isFullDbSearch      = true;
    m_exit                = false;
}

CoverThread::~CoverThread()
{
}


//! -- CoverThread::setRequest -------------------------------------------------
void CoverThread::setRequest(const QString& artist, const QString& album)
{
    m_list_param.clear();
    
    RequestParam param = RequestParam();        
    param.artist     = artist;
    param.album      = album;
    param.cover_path = MEDIA::getCoverName(artist, album);
        
    const QString path = UTIL::CONFIGDIR + "/albums/" + param.cover_path;

    if(!QFile::exists(path)) {
      m_list_param.append (param);
    }        
    m_isFullDbSearch = false;
}
   

//! -- CoverThread::run --------------------------------------------------------
void CoverThread::run()
{
    Debug::debug() << "  ---- CoverThread -> run";
    if(m_isFullDbSearch)
      slot_check_database();
    else 
      // take request as parameter
      
    m_max = m_list_param.size();
    
    slot_process_cover_search();

    //! Start thread event loop --> makes signals and slots work
    exec();
    Debug::debug() <<"  ---- CoverThread -> exited";
    
    m_isFullDbSearch      = true;
    m_list_param.clear();  
}

//! -- slot_finish_cover_search ------------------------------------------------
void CoverThread::slot_finish_cover_search()
{
    //Debug::debug() << " ---- CoverThread -> finished";

    //! deletes previous providers 
    qDeleteAll(m_coverprovider_list);
    m_coverprovider_list.clear();
    
    emit finished();
    exit();
}

//! -- slot_process_cover_search -----------------------------------------------
void CoverThread::slot_process_cover_search() 
{
    Debug::debug() << " ---- CoverThread -> slot_process_cover_search";    
    if(!m_list_param.isEmpty() && !m_exit)
    {
      emit progress ( ((m_max - m_list_param.size())*100)/m_max );
      
      m_current_param = m_list_param.takeFirst();
             
      
      //! set providers 
      m_coverprovider_list.append( new CoverProvider_LastFm(0) );
      m_coverprovider_list.append( new CoverProvider_Discogs(0) );

      foreach (CoverProvider* provider, m_coverprovider_list) {
        connect(provider, SIGNAL(search_finished(int)), this, SLOT(slot_start_cover_provider()));
        connect(provider, SIGNAL(data_received(int,QByteArray)), this, SLOT(slot_save_cover(int,QByteArray)));
      }
    
      //! start cover provider
      m_current_provider_id = 0;
      slot_start_cover_provider();
    }
    else 
    {
      slot_finish_cover_search();
    }
}


//! -- slot_start_cover_provider -----------------------------------------------
void CoverThread::slot_start_cover_provider()
{
    if(m_current_provider_id < m_coverprovider_list.size() ) 
    {
      CoverProvider* provider = m_coverprovider_list.at(m_current_provider_id);
      
      provider->set_search(m_current_param.artist, m_current_param.album); 
      provider->start_search(++m_current_request_id);

      m_current_provider_id++;
    }
    else 
    {
      //! process next request
      slot_process_cover_search();    
    }
}



//! -- slot_check_database -----------------------------------------------------
void CoverThread::slot_check_database()
{
    //Debug::debug() << " --- CoverThread -> slot_check_database";      
    Database db;
    if (!db.connect("CoverThread")) 
      return;
      
    m_list_param.clear();
      
    //! albums database loop
    QSqlQuery query("SELECT name,cover,artist_name FROM view_albums",*db.sqlDb());
    while (query.next()) 
    {
        RequestParam param = RequestParam();
        
        param.artist     = query.value(2).toString();  // artist
        param.album      = query.value(0).toString();  // album
        param.cover_path = query.value(1).toString();  // covername

        const QString path = UTIL::CONFIGDIR + "/albums/" + param.cover_path;

        if(!QFile::exists(path)) {
          m_list_param.append (param);
        }
    } // fin while album query
}


//! -- slot_save_cover ---------------------------------------------------------
void CoverThread::slot_save_cover(int request_id, QByteArray bytes) 
{
    Debug::debug() << " ---- CoverThread -> slot_save_cover";
    if(m_current_request_id != request_id) {
      Debug::debug() << " ---- CoverThread -> bad request id";
      slot_start_cover_provider();  
      return;
    }
  
    //! filepath for cover file
    const QString filePath = UTIL::CONFIGDIR + "/albums/" + m_current_param.cover_path;
    
    QFile file(filePath);
    if(file.exists()) 
    {
      Debug::debug() << " ---- CoverThread -> file exists, start next cover search";
    }
    else if (!file.open(QIODevice::WriteOnly)) 
    {
      Debug::debug() << " ---- CoverThread -> error writing album image";
    }
    else 
    {
      QImage image = QImage::fromData(bytes);
      image = image.scaled(QSize(110, 110), Qt::KeepAspectRatio, Qt::SmoothTransformation);
      image.save(filePath, "png", -1);
    }

    //! process next request
    slot_process_cover_search();
}
