/* YaRock Player
* Copyright (C) 2010-2013 Sebastien amardeilh (sebastien.amardeilh+yarock@gmail.com)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

//! local
#include "navigatorbar.h"
#include "menuitem.h"
#include "menumodel.h"

#include "yarocksettings.h"
//#include "debug.h"


//! Qt
#include <QPainter>

#include <QStyle>
#include <QCleanlooksStyle>
#include <QWindowsStyle>
#include <QMenu>

/*
********************************************************************************
*                                                                              *
*    Class NavigatorBar                                                        *
*                                                                              *
********************************************************************************
*/
NavigatorBar::NavigatorBar(QWidget * parent) : QToolBar(parent)
{
    this->setContentsMargins(0, 0, 0, 0);
    this->setSizePolicy( QSizePolicy::Minimum, QSizePolicy::Minimum );
    this->setToolButtonStyle( Qt::ToolButtonTextBesideIcon );
    this->setIconSize( QSize( 20, 20) );
 
    //!  Remove that ugly frame panel around the toolbar for Cleanlooks & Windows styles.
    if (qobject_cast <QCleanlooksStyle*>(this->style()) || qobject_cast <QWindowsStyle*>(this->style()))
      setStyleSheet("QToolBar { border: 0px none black; }");
}

//! ----- NavigatorBar::setModel -----------------------------------------------
void NavigatorBar::setModel(MenuModel * m) 
{
    m_model = m;
    connect(m_model, SIGNAL(modelUpdated()),this, SLOT(updateEntry()));
    updateEntry();
}

//! ----- NavigatorBar::updateEntry --------------------------------------------
void NavigatorBar::updateEntry() 
{
    QString text;
    QIcon   icon;
    this->clear();
    
    NavigatorItemButton *homeButton = new NavigatorItemButton(QIcon(":/images/go-home.png"),tr("Home") , this);

    this->addWidget(homeButton);
    m_listButton.append(homeButton);
    
    for (int i=0; i < m_model->rowCount(QModelIndex()); i++) 
    {
      const QModelIndex &idx = m_model->index(i, 0, QModelIndex());

      if( !idx.isValid() ) continue;
      
      //! build button
      if(m_model->rowCount(idx) == 0) 
      {
        QAction *a = idx.data(MenuActionRole).value<QAction*>();
        //homeMenu->addAction(a);
        homeButton->itemMenu()->addAction(a);
      }
      else 
      {
        text   = idx.data(Qt::DisplayRole).toString();
        icon   = qvariant_cast<QIcon>(idx.data(Qt::DecorationRole));

        NavigatorItemButton *button = new NavigatorItemButton(icon,text , this);

        this->addWidget(button);
        m_listButton.append(button);
        
        //! build menu
        for (int j=0; j < m_model->rowCount(idx); j++) 
        {
            const QModelIndex childIdx = m_model->index(j, 0, idx);
            QAction *a = childIdx.data(MenuActionRole).value<QAction*>();
            button->itemMenu()->addAction(a);
        }
      }
    }
   
    if(SETTINGS()->_popup_on_mouseover) {
      foreach(NavigatorItemButton* button, m_listButton) {
        connect(button, SIGNAL(menu_activated()), this, SLOT(hideMenu())); 
      }
    }
      
    update();
    updateGeometry();
}



QSize NavigatorBar::sizeHint() const
{
    int l, t, r, b;
    getContentsMargins(&l, &t, &r, &b);

    int width = 0;
    foreach(NavigatorItemButton* button, m_listButton) {
      width += button->width() + 10;
    }
    
    return QSize( l + r + width, t + b + iconSize().height());
}

void NavigatorBar::hideMenu ( )
{
    NavigatorItemButton *b = qobject_cast<NavigatorItemButton *>(sender());
    foreach(NavigatorItemButton* button, m_listButton) {
     if(button != b)
       button->itemMenu()->hide();
    }

}


/*
********************************************************************************
*                                                                              *
*    Class NavigatorItemButton                                                 *
*                                                                              *
********************************************************************************
*/
NavigatorItemButton::NavigatorItemButton( const QIcon &icon, const QString &text, QWidget *parent )
    : QPushButton( icon, "", parent )
{
    is_popup_mouse_over = SETTINGS()->_popup_on_mouseover;
    
    if(!is_popup_mouse_over)
      this->setFocusPolicy(Qt::ClickFocus);
    else
      this->setFocusPolicy(Qt::TabFocus);
    
    this->setIconSize( QSize(20,20) );
    this->setFlat ( true );

    bool option_to_show_menu_text = false;
    if(option_to_show_menu_text)
      this->setText(text);
    else
      this->setToolTip(text);
    
    m_font  = QFont("Segoe UI",9, QFont::Normal);
    m_menu  = new NavigatorItemMenu(this);
    
    if(!is_popup_mouse_over)
      connect(this, SIGNAL(clicked(bool)), this, SLOT(slot_button_clicked()));
}




void NavigatorItemButton::paintEvent( QPaintEvent* event )
{
    QPushButton::paintEvent(event);
}



QSize NavigatorItemButton::sizeHint() const
{
    QSize size = QPushButton::sizeHint();
    
    int width = 8;
    if( !icon().isNull() ) {
        width += iconSize().width();
    }
    
    if( !text().isEmpty() )  {
      
        QFontMetrics fm( m_font );
        width += fm.width( text() );
        width += 10; // paddding
    }
    size.setWidth( width );
    return size;
}


void NavigatorItemButton::enterEvent(QEvent *e)
{
    if(is_popup_mouse_over)
      slot_button_clicked();
    
    QPushButton::enterEvent(e);    
}

void NavigatorItemButton::leaveEvent(QEvent *e)
{
    QPushButton::leaveEvent(e);
}

void NavigatorItemButton::slot_button_clicked()
{
    //Debug::debug() << "--- NavigatorItemButton -> slot_button_clicked ";

    if(!m_menu->isVisible()) {
      QPoint location = this->mapToGlobal(QPoint(0,35));
      m_menu->popup(location);
      m_menu->show();
      emit menu_activated();
    }
}

void NavigatorItemButton::mousePressEvent ( QMouseEvent * e )
{
    if(is_popup_mouse_over)
      e->ignore();  
    else
      QPushButton::mousePressEvent(e);
}



    
/*
********************************************************************************
*                                                                              *
*    Class NavigatorItemMenu                                                   *
*                                                                              *
********************************************************************************
*/  
NavigatorItemMenu::NavigatorItemMenu(QWidget * parent) :  QMenu(parent)
{
    QFont normalFont = QFont("Segoe UI", 9, QFont::Normal);
    setFont(normalFont);
}

void NavigatorItemMenu::enterEvent(QEvent *e)
{
Q_UNUSED(e)  
}

void NavigatorItemMenu::leaveEvent(QEvent *e)
{
Q_UNUSED(e)  
    /* si le bouton n'a pas le focus on cache le menu, cela evite 
       d'attendre un click souris sur autre chose */ 
    if( !this->parentWidget()->underMouse() ) {
      this->hide();    
    }   
}

