/* YaRock Player
* Copyright (C) 2010-2013 Sebastien amardeilh (sebastien.amardeilh+yarock@gmail.com)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "provider_discogs.h"
#include "networkaccess.h"
#include "debug.h"

// QJson
#include <qjson/parser.h>


/*
********************************************************************************
*                                                                              *
*    Class DiscogsProvider                                                     *
*                                                                              *
********************************************************************************
*/
DiscogsProvider::DiscogsProvider(QObject* parent)  : InfoProvider("discogs")
{
    m_max_release  = 15;
}

DiscogsProvider::~DiscogsProvider()
{
    cancel_searches();
}

/*******************************************************************************
  DiscogsProvider::init
*******************************************************************************/ 
void DiscogsProvider::init()
{

}

/*******************************************************************************
  DiscogsProvider::cancel_searches
*******************************************************************************/ 
void DiscogsProvider::cancel_searches() 
{
   /* Debug::debug() << "Discogs: cancel_searches";
    foreach (QObject* object, m_list_requests.keys()) {
      m_list_requests.take(object);
      delete object;
    }
    m_list_requests.clear();*/
  
    /* clear artist */
    m_disco_artist.m_name.clear();
    m_disco_artist.m_discogs_url.clear();
    m_disco_artist.m_releases.clear();
    
    /* clear album */
    m_map_request_album.clear();
}


/*******************************************************************************
  DiscogsProvider::fetch_info
*******************************************************************************/ 
void DiscogsProvider::fetch_info(int id) 
{
    if( !is_enabled() ) return;
    fetch_artist( id, m_current_media->data.at(TRACK::artist).toString() );
    //fetch_release( id, m_current_media->data.at(TRACK::album).toString() );
}


/*******************************************************************************
  DiscogsProvider::fetch_artist
*******************************************************************************/ 
void DiscogsProvider::fetch_artist(int id, const QString& artist_name)
{
    QUrl url("http://api.discogs.com/artist/" + artist_name);
    url.addQueryItem( "api_key", DISCOGS_API_KEY );    
    url.addQueryItem( "page", QString::number( 1 ) );    
    url.addQueryItem( "releases", "1" );
    url.addQueryItem( "f", "json" );
        
    Debug::debug() << "Discogs::get_artist " << url;
    QObject* reply = HTTP()->get(url);
    m_list_requests[reply] = id;
        
    connect(reply, SIGNAL(data(QByteArray)), this, SLOT(slot_parse_artist_response(QByteArray)));
    connect(reply, SIGNAL(error(QNetworkReply*)), this, SLOT(slot_artist_search_stop()));
}

/*******************************************************************************
  DiscogsProvider::slot_artist_search_stop
*******************************************************************************/ 
void DiscogsProvider::slot_artist_search_stop()  
{
    Debug::debug() << "Discogs::slot_artist_search_stop";
    
    // Get id from sender reply
    QObject* reply = qobject_cast<QObject*>(sender());
    if (!reply || !m_list_requests.contains(reply))   return;    
    
    const int id = m_list_requests.take(reply);
    
    // Send signal for data reception
    emit search_artist_finished(id);
}    

/*******************************************************************************
  DiscogsProvider::slot_parse_artist_response
*******************************************************************************/ 
void DiscogsProvider::slot_parse_artist_response(QByteArray bytes) 
{
    Debug::debug() << "Discogs::slot_parse_artist_response";
    
    /*-------------------------------------------------*/
    /* Get id from sender reply                        */
    /* ------------------------------------------------*/
    QObject* reply = qobject_cast<QObject*>(sender());
    if (!reply || !m_list_requests.contains(reply))   return;    
    const int id = m_list_requests.take(reply);

    /*-------------------------------------------------*/
    /* Parse response                                  */
    /* ------------------------------------------------*/
    QJson::Parser parser;
    bool ok;
    QVariantMap reply_map = parser.parse(bytes, &ok).toMap();

    if (!ok || !reply_map.contains("resp")) {
      Debug::debug() << " Discogs -> no artist id found";    
      emit search_artist_finished(id);
      return;
    }
      
    //foreach (QString key, reply_map.keys() )
    //   Debug::debug() << " Discogs -> value" << reply_map.value(key);
      
    QVariantMap resp_map =  qvariant_cast<QVariantMap>(reply_map.value("resp"));

    //foreach (QString key, resp_map.keys()) {
    //  Debug::debug() << "resp_map key = " << key;
    //}

    if (resp_map.contains("artist")) 
    {
       m_disco_artist = DISCOGS::Artist();
      
       QVariantMap artist_map =  qvariant_cast<QVariantMap>(resp_map.value("artist"));

       foreach (QString key, artist_map.keys())
       {
           //Debug::debug() << "artist_map key = " << key;
           //Debug::debug() << "artist_map value = " << artist_map.value(key);
           if(key == "resource_url")
             m_disco_artist.m_discogs_url = artist_map.value(key).toString();

           if(key == "id")
             m_disco_artist.m_id = artist_map.value(key).toInt();
           else if (key == "uri")
             m_disco_artist.m_discogs_url = artist_map.value(key).toString();
           else if  (key == "name")
             m_disco_artist.m_name = artist_map.value(key).toString();
           else if(key == "releases")
           {
              QVariantList releases = artist_map.value(key).toList();

              int count = 0;
              foreach (const QVariant& release, releases)
              {
                  QVariantMap release_map = release.toMap();

                  if (release_map.contains("type"))
                    if (release_map.value("type").toString() != "master")
                      continue;


                  DISCOGS::Release album = DISCOGS::Release();      
                  album.m_id        = release_map.value("id").toInt();
                  album.m_thumb_url = release_map.value("thumb").toString();      
                  album.m_title     = release_map.value("title").toString();      
                  album.m_year      = release_map.value("year").toString();   
                  if(count++ >= m_max_release)
                    break;
                  
//                   Debug::debug() << "release_map id     " << release_map.value("id").toInt();
//                   Debug::debug() << "release_map thumb  " << release_map.value("thumb").toString();
//                   Debug::debug() << "release_map title  " << release_map.value("title").toString();
//                   Debug::debug() << "release_map year   " << release_map.value("year").toString();

                  m_disco_artist.m_releases.append(album);
                  
                  QUrl album_image_url = QUrl(album.m_thumb_url);
                  if(album_image_url.isValid()) {
                    QObject* reply = HTTP()->get( QUrl(album.m_thumb_url) );
                    m_list_requests[reply]  = id;
                    m_map_request_album[reply] = album.m_id;
                    connect(reply, SIGNAL(data(QByteArray)), this, SLOT(slot_cover_received(QByteArray)));
                  }
              } // end release loop
           }

       }
    }
    
    //Debug::debug() << " Discogs m_disco_artist->m_id" << m_disco_artist.m_id;
    //Debug::debug() << " Discogs m_disco_artist->m_name" << m_disco_artist.m_name;
    //Debug::debug() << " Discogs m_disco_artist->m_discogs_url" << m_disco_artist.m_discogs_url;
    
    // Send signal for data reception
    emit artist_data_received(id, m_disco_artist);
}  

/*******************************************************************************
  DiscogsProvider::slot_cover_received
*******************************************************************************/ 
void DiscogsProvider::slot_cover_received(QByteArray bytes)
{
    //Debug::debug() << " ---- Discogs -> slot_cover_received";
    
    // Get id from sender reply
    QObject* reply = qobject_cast<QObject*>(sender());
    if (!reply || !m_list_requests.contains(reply))   return;    
    
    const int id = m_list_requests.take(reply);

    if(m_map_request_album.contains(reply))
    {
      //Debug::debug() << " ---- Discogs -> slot_cover_received album found";
      int release_id = m_map_request_album.take(reply);
      emit release_cover_received(id, release_id, bytes);
    }    
}

/*******************************************************************************
  DiscogsProvider::fetch_release
*******************************************************************************/ 
void DiscogsProvider::fetch_release(int id, const QString& release_name )
{
  //! TODO

}

