/* YaRock Player
* Copyright (C) 2010-2013 Sebastien amardeilh (sebastien.amardeilh+yarock@gmail.com)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// local
#include "Database/databasebuilder.h"
#include "Database/database.h"
#include "Database/databasemanager.h"  // 0.0.52

#include "MediaItem/mediaitem.h"
#include "MediaItem/playlist_parser.h"

#include "utilities.h"                 // CONFIGDIR
#include "debug.h"

// taglib
#include <taglib/fileref.h>
#include <taglib/tag.h>
#include <taglib/tstring.h>
#include <taglib/mpegfile.h>

// Qt
#include <QtCore>
#include <QImage>
#include <QtSql/QSqlDatabase>
#include <QtSql/QSqlQuery>
#include <QtSql/QSqlDriver>
#include <QtSql/QSqlError>

#include <QFuture>
#include <QFutureWatcher>
#include <QtConcurrentRun>


const QStringList mediaFilter = QStringList() 
         /* Audio */    << "*.mp3"  << "*.ogg" << "*.wav" << "*.flac" << "*.m4a" << "*.aac"
         /* Playlist */ << "*.m3u" << "*.m3u8" << "*.pls" << "*.xspf";
 
/*******************************************************************************
    concurrent fonction
*******************************************************************************/
int findFileCount(const QStringList _folders)
{
    Debug::debug() << "- DataBaseBuilder -> findFileCount ";

    int count = 0;
  
    //! analyse des repertoire de la collection
    foreach(const QString& folder, _folders ) {
      QDir dirlist(folder);
      dirlist.setNameFilters(mediaFilter);
      QDirIterator dirIterator(dirlist,QDirIterator::Subdirectories);
      while (dirIterator.hasNext()) {
         dirIterator.next();
         count++;
      }
    }  
  
    return count;
}

 
/*
********************************************************************************
*                                                                              *
*    Class DataBaseBuilder                                                     *
*                                                                              *
********************************************************************************
*/ 
DataBaseBuilder::DataBaseBuilder()
{
    _exit            = false;
   
}


//! --------- DataBaseBuilder::setFileCount ------------------------------------
void DataBaseBuilder::setFileCount()
{
    QFutureWatcher<int>* watcher = static_cast<QFutureWatcher<int>*>(sender());
    watcher->deleteLater();

    int i = watcher->result();
    Debug::debug() << "- DataBaseBuilder -> setFileCount " << i;
  
    _fileCount = i;
}




//! --------- DataBaseBuilder::rebuildFolder -----------------------------------
// User entry point -> add folder to parse
void DataBaseBuilder::rebuildFolder(QStringList folder)
{
    _folders.append(folder);
    
    _fileCount     = -1;

    QFuture<int> future = QtConcurrent::run(&findFileCount,_folders);
    QFutureWatcher<int>* watcher = new QFutureWatcher<int>(this);
    watcher->setFuture(future);
    connect(watcher, SIGNAL(finished()), SLOT(setFileCount()));
}

//! --------- DataBaseBuilder::run --------------------------------------------- 
void DataBaseBuilder::run()
{
    if (_folders.isEmpty()) return;
    int percent    = 0;
    int idxCount   = 0;

    
    
    Database db;
    if (!db.connect("DataBaseBuilder")) return;    
    m_sqlDb = db.sqlDb();
    
    Debug::debug() << "- DataBaseBuilder -> starting Database update";

    //! Populate dbAudioFiles map with all tracks from database
    QSqlQuery trackQuery("SELECT filename, mtime FROM tracks;",*m_sqlDb);
    while (trackQuery.next())
      dbMediaFiles.insert(trackQuery.value(0).toString(),trackQuery.value(1).toUInt());
    
    //! Populate dbPlaylistFiles map with "User type" Playlist from database
    QSqlQuery playlistQuery("SELECT filename, mtime FROM playlists WHERE type=1;",*m_sqlDb);
    while (playlistQuery.next())
      dbMediaFiles.insert(playlistQuery.value(0).toString(),playlistQuery.value(1).toUInt());
    
    QFileInfo fileInfo;

    //On SQLite --> it's MUCH faster to have everything in one transaction 
    //with only one disk write than to commit every each insert individually
    QSqlQuery("BEGIN TRANSACTION;",*m_sqlDb);

    do 
    {
        QDir dirlist(_folders.takeFirst());
        dirlist.setNameFilters(mediaFilter);

        QDirIterator dirIterator(dirlist,QDirIterator::Subdirectories);
        while (dirIterator.hasNext()) {
            dirIterator.next();
            fileInfo = dirIterator.fileInfo();
            
            //! skip non file
            if (!fileInfo.isFile()) continue;
                    
            //! If the file is NOT in database then insert
            if (!dbMediaFiles.contains(fileInfo.filePath().toUtf8())) 
            {
               if (MEDIA::isAudioFile(fileInfo.filePath().toUtf8()) )           
                 insertTrack(fileInfo.absoluteFilePath());
               else
                 insertPlaylist(fileInfo.absoluteFilePath());
            }
            //! If the file is in database but has another mtime then update it
            else if (dbMediaFiles[fileInfo.filePath().toUtf8()]!=fileInfo.lastModified().toTime_t()) 
            {
               if (MEDIA::isAudioFile(fileInfo.filePath().toUtf8()) )           
                 updateTrack(fileInfo.absoluteFilePath());
               else
                 updatePlaylist(fileInfo.absoluteFilePath());
            }
                
            dbMediaFiles.remove(fileInfo.filePath().toUtf8());

            //! signal progress
            if(_fileCount!=-1) {
              percent = 100 - ((_fileCount - ++idxCount) * 100 / _fileCount);
              emit buildingProgress(percent);  
            }
    
        } // WHILE dirIterator.hasNext
    } while (_folders.size()>0 && !_exit);

    //! Get files that are in DB but not on harddrive
    QHashIterator<QString, uint> i(dbMediaFiles);
    while (i.hasNext()) {
        i.next();
        if( MEDIA::isAudioFile(i.key()) )
          removeTrack(i.key());
        else
          removePlaylist(i.key());
    }
       
    dbMediaFiles.clear();

    // Check for interprets/albums/genres... that are not used anymore
    cleanUpDatabase();

    // Now write all data to the disk
    QSqlQuery("COMMIT TRANSACTION;",*m_sqlDb);


    Debug::debug() << "- DataBaseBuilder -> end Database update";
    if(!_exit)
      emit buildingFinished();
}



//! --------- DataBaseBuilder::insertTrack -------------------------------------
//   -     MEDIA::FromLocalFile(fname) to get track metadata
//   -     MEDIA::getCoverName(media) to get hash of covername
void DataBaseBuilder::insertTrack(const QString& filename)
{
    QFileInfo fileInfo(filename);

    QString fname = fileInfo.filePath().toUtf8();

    uint mtime = QFileInfo(filename).lastModified().toTime_t();

    Debug::debug() << "- DataBaseBuilder -> insert track :" << filename;

    //! storage localtion
    QString storageLocation = UTIL::CONFIGDIR + "/albums/";

    //! Read tag from URL file (with taglib) 
    int disc_number = 0;
    MediaItemPtr media    = MEDIA::FromLocalFile(fname, &disc_number);
    QString  cover_name   = MEDIA::getCoverName(media);
    

    //! GENRE part in database
    int id_genre = insertGenre( media->data.at(TRACK::genre).toString() );
    
    //! YEAR part in database
    int id_year = insertYear( QString::number(media->data.at(TRACK::year).toUInt()) );
    
    //! ARTIST part in database
    int id_artist = insertArtist( media->data.at(TRACK::artist).toString() );
    
    //! ALBUM part in database
    int id_album = insertAlbum(
        media->data.at(TRACK::album).toString(),
        id_artist,
        cover_name,
        QString::number(media->data.at(TRACK::year).toUInt()),
        disc_number        
        );
    
    //! mise à jour du cover
    storeCoverArt(storageLocation + cover_name, media->data.at(TRACK::url).toString());

    if( DatabaseManager::instance()->DB_PARAM().checkCover )
      recupCoverArtFromDir(storageLocation + cover_name, media->data.at(TRACK::url).toString());

    //! TRACK part in database
    QSqlQuery query(*m_sqlDb);
    query.prepare("INSERT INTO `tracks`(`filename`,`trackname`,`number`,`length`,`artist_id`,`album_id`,`year_id`,`genre_id`,`mtime`,`playcount`,`rating`,`albumgain`,`albumpeakgain`,`trackgain`,`trackpeakgain`)" \
                  "VALUES(?,?,?,?,?,?,?,?,?,?,?,?,?,?,?);");
    query.addBindValue(fname);
    query.addBindValue(media->data.at(TRACK::title).toString());
    query.addBindValue(media->data.at(TRACK::num).toUInt());
    query.addBindValue(media->data.at(TRACK::duration).toInt());
        
    query.addBindValue(id_artist);
    query.addBindValue(id_album);
    query.addBindValue(id_year);
    query.addBindValue(id_genre);
        
    query.addBindValue(mtime);
    query.addBindValue(media->data.at(TRACK::playcount).toInt());
    query.addBindValue(media->data.at(TRACK::rating).toDouble());
    query.addBindValue(media->data.at(TRACK::albumGain).toDouble());
    query.addBindValue(media->data.at(TRACK::albumPeak).toDouble());
    query.addBindValue(media->data.at(TRACK::trackGain).toDouble());
    query.addBindValue(media->data.at(TRACK::trackPeak).toDouble());
    query.exec();
    
    // delete media from memory    
    if(media)
      media.reset();
}

int DataBaseBuilder::insertGenre(const QString & genre)
{
    QSqlQuery q("", *m_sqlDb);
    q.prepare("SELECT `id` FROM `genres` WHERE `genre`=:val;");
    q.bindValue(":val", genre );
    q.exec();

    if ( !q.next() ) {
      q.prepare("INSERT INTO `genres`(`genre`) VALUES (:val);");
      q.bindValue(":val", genre);
      q.exec();
              
      if(q.numRowsAffected() < 1) return -1;
      q.prepare("SELECT `id` FROM `genres` WHERE `genre`=:val;");
      q.bindValue(":val", genre);
      q.exec();
      q.next();        
    }
    return q.value(0).toString().toInt();
}    
    
int DataBaseBuilder::insertYear(const QString & year)
{
    QSqlQuery q("", *m_sqlDb);
    q.prepare("SELECT `id` FROM `years` WHERE `year`=:val;");
    q.bindValue(":val", year );
    q.exec();

    if ( !q.next() ) {
      q.prepare("INSERT INTO `years`(`year`) VALUES (:val);");
      q.bindValue(":val", year);
      q.exec();
              
      if(q.numRowsAffected() < 1) return -1;
      q.prepare("SELECT `id` FROM `years` WHERE `year`=:val;");
      q.bindValue(":val", year);
      q.exec();
      q.next();        
    }
    return q.value(0).toString().toInt();
}

int DataBaseBuilder::insertArtist(const QString & artist)
{
    QSqlQuery q("", *m_sqlDb);
    q.prepare("SELECT `id` FROM `artists` WHERE `name`=:val;");
    q.bindValue(":val", artist );
    q.exec();

    if ( !q.next() ) {
      q.prepare("INSERT INTO `artists`(`name`,`favorite`,`playcount`,`rating`) VALUES (:val,0,0,-1);");
      q.bindValue(":val", artist);
      q.exec();
              
      if(q.numRowsAffected() < 1) return -1;
      q.prepare("SELECT `id` FROM `artists` WHERE `name`=:val;");
      q.bindValue(":val", artist);
      q.exec();
      q.next();        
    }
    return q.value(0).toString().toInt();
}


int DataBaseBuilder::insertAlbum(const QString & album, int artist_id,const QString & cover,const QString & year,int disc)
{
    QSqlQuery q("", *m_sqlDb);
    q.prepare("SELECT `id` FROM `albums` WHERE `name`=:val AND `artist_id`=:id AND `disc`=:dn;");
    q.bindValue(":val", album );
    q.bindValue(":id", artist_id );
    q.bindValue(":dn", disc );
    q.exec();

    if ( !q.next() ) {
      q.prepare("INSERT INTO `albums`(`name`,`artist_id`,`cover`,`year`,`favorite`,`playcount`,`rating`,`disc`) VALUES (:val,:id,:cov,:y,0,0,-1,:dn);");
      q.bindValue(":val", album);
      q.bindValue(":id", artist_id );
      q.bindValue(":cov", cover );
      q.bindValue(":y", year );
      q.bindValue(":dn", disc );      
      q.exec();
              
      if(q.numRowsAffected() < 1) return -1;
      q.prepare("SELECT `id` FROM `albums` WHERE `name`=:val AND `artist_id`=:id AND `disc`=:dn;");
      q.bindValue(":val", album );
      q.bindValue(":id", artist_id );
      q.bindValue(":dn", disc );
      q.exec();
      q.next();        
    }
    return q.value(0).toString().toInt();
}



//! --------- DataBaseBuilder::storeCoverArt -----------------------------------
void DataBaseBuilder::storeCoverArt(const QString& coverFilePath, const QString& trackFilename)
{
    //Debug::debug() << "- DataBaseBuilder -> storeCoverArt " << coverFilePath;

    //! check if coverArt already exist
    QFile file(coverFilePath);
    if(file.exists()) return;
  
    //! get cover from file
    TagLib::MPEG::File mpegFile(QFileInfo(trackFilename).filePath().toLocal8Bit());
    TagLib::ID3v2::Tag *id3tag = mpegFile.ID3v2Tag(false);
    if (id3tag) {
      TagLib::ID3v2::AttachedPictureFrame *selectedFrame = MEDIA::attachedPictureFrame(id3tag);
    
      if (selectedFrame)
      {
        QByteArray pictureData = QByteArray(selectedFrame->picture().data(), selectedFrame->picture().size());
        if(!pictureData.isNull()) 
        {
          if (!file.open(QIODevice::WriteOnly)) 
          {
            Debug::debug() << "- DataBaseBuilder Error opening file for writing" << file.fileName();
          }
          else {
            Debug::debug() << "- DataBaseBuilder -> new cover file stored " << coverFilePath;
            //! WARNING la taille est fixée à 110 par 110 (taille utile à la vue album)
            QImage image = QImage::fromData(pictureData);
            image = image.scaled(QSize(110, 110), Qt::KeepAspectRatio, Qt::SmoothTransformation);
            image.save(coverFilePath, "png", -1);
          }
        }
      }
    }
}

//! --------- DataBaseBuilder::recupCoverArtFromDir ----------------------------
void DataBaseBuilder::recupCoverArtFromDir(const QString& coverFilePath, const QString& trackFilename)
{
    //Debug::debug() << "- DataBaseBuilder -> recupCoverArtFromDir " << coverFilePath;

    //! check if coverArt already exist
    QFile file(coverFilePath);
    if(file.exists()) return ;
  
    //! search album art into file source directory
    const QStringList imageFilters = QStringList() << "*.jpg" << "*.png";
    QDir sourceDir(QFileInfo(trackFilename).absolutePath());
  
    sourceDir.setNameFilters(imageFilters);
  
    QStringList entryList = sourceDir.entryList(imageFilters, QDir::Files, QDir::Size);
 
    while(!entryList.isEmpty()) {
      //! I take the first one (the biggest one)
      //!WARNING simplification WARNING
      QString file = QFileInfo(trackFilename).absolutePath() + "/" + entryList.takeFirst();
      QImage image = QImage(file);
      //! check if not null image (occur when file is KO)
      if(!image.isNull()) {
        image = image.scaled(QSize(110, 110), Qt::KeepAspectRatio, Qt::SmoothTransformation);
        //! check if save is OK
        if(image.save(coverFilePath, "png", -1))
          break;
      }
    }
}


//! --------- DataBaseBuilder::updateTrack -------------------------------------
void DataBaseBuilder::updateTrack(const QString& filename)
{
    removeTrack(filename);
    insertTrack(filename);
}


//! --------- DataBaseBuilder::removeTrack -------------------------------------
void DataBaseBuilder::removeTrack(const QString& filename)
{
    Debug::debug() << "- DataBaseBuilder -> Deleting track :" << filename;
    QFileInfo fileInfo(filename);
    QString fname = fileInfo.filePath().toUtf8();

    QSqlQuery query(*m_sqlDb);
    query.prepare("DELETE FROM `tracks` WHERE `filename`=?;");
    query.addBindValue(fname);
    query.exec();
}

//! --------- DataBaseBuilder::cleanUpDatabase ---------------------------------
void DataBaseBuilder::cleanUpDatabase()
{
    {
      QSqlQuery query("DELETE FROM `albums` WHERE `id` NOT IN (SELECT `album_id` FROM `tracks` GROUP BY `album_id`);", *m_sqlDb);
    }
    {
      QSqlQuery query("DELETE FROM `genres` WHERE `id` NOT IN (SELECT `genre_id` FROM `tracks` GROUP BY `genre_id`);", *m_sqlDb);
    }
    {      
      QSqlQuery query("DELETE FROM `artists` WHERE `id` NOT IN (SELECT `artist_id` FROM `tracks` GROUP BY `artist_id`);", *m_sqlDb);
    }
    {
      QSqlQuery query("DELETE FROM `years` WHERE `id` NOT IN (SELECT `year_id` FROM `tracks` GROUP BY `year_id`);", *m_sqlDb);
    }
    {
      QSqlQuery query("DELETE FROM `playlist_items` WHERE `playlist_id` NOT IN (SELECT `id` FROM `playlists`);", *m_sqlDb);
    }
}


//! --------- DataBaseBuilder::insertPlaylist ----------------------------------
void DataBaseBuilder::insertPlaylist(const QString& filename)
{
    QFileInfo fileInfo(filename);
    QString fname = fileInfo.filePath().toUtf8();
    QString pname = fileInfo.baseName();
    uint mtime    = fileInfo.lastModified().toTime_t();
    
    Debug::debug() << "- DataBasePlsBuilder -> insert playlist :" << filename;

    int favorite = 0;
    
    QSqlQuery query(*m_sqlDb);
    query.prepare("INSERT INTO `playlists`(`filename`,`name`,`type`,`favorite`,`mtime`)" \
                  "VALUES(?," \
                  "       ?," \
                  "       ?," \
                  "       ?," \
                  "       ?);");
    query.addBindValue(fname);
    query.addBindValue(pname);
    query.addBindValue((int) T_FILE);
    query.addBindValue(favorite);
    query.addBindValue(mtime);
    query.exec();
    

    QList<MediaItemPtr> list =  MEDIA::PlaylistFromFile(filename);
    foreach (MediaItemPtr mi, list) 
    {
      QString url           = mi->data.at(TRACK::url).toString();
      QString itemUrlName   = url; //QFileInfo(url).filePath().toUtf8();
      QString itemName      = QFileInfo(url).baseName();

      //! Playlist Item part in database
      Debug::debug() << "- DataBasePlsBuilder -> insert playlistitem itemUrlName:" << itemUrlName;

      query.prepare("INSERT INTO `playlist_items`(`url`,`name`,`playlist_id`)" \
                    "VALUES(?," \
                    "       ?," \
                    "       (SELECT `id` FROM `playlists` WHERE `filename`=?));");
      query.addBindValue(itemUrlName);
      query.addBindValue(itemName);
      query.addBindValue(fname);
      query.exec();
    } // end foreach url into playlist
}


//! --------- DataBaseBuilder::updatePlaylist ----------------------------------
void DataBaseBuilder::updatePlaylist(const QString& filename)
{
    removePlaylist(filename);
    insertPlaylist(filename);
}


//! --------- DataBaseBuilder::removePlaylist ----------------------------------
void DataBaseBuilder::removePlaylist(const QString& filename)
{
    Debug::debug() << "- DataBasePlsBuilder -> Deleting playlist :" << filename;
    QFileInfo fileInfo(filename);
    QString fname = fileInfo.filePath().toUtf8();

    QSqlQuery query(*m_sqlDb);
    query.prepare("DELETE FROM `playlists` WHERE `filename`=?;");
    query.addBindValue(fname);
    query.exec();
}


