/* YaRock Player
* Copyright (C) 2010-2013 Sebastien amardeilh (sebastien.amardeilh+yarock@gmail.com)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "Dbus/dbusnotification.h"
#include "EnginePlayer/engineplayer.h"
#include "MediaItem/mediaitem.h"
#include "constants.h"
#include "mainwindow.h"

#include "yarocksettings.h"
#include "debug.h"


#include <QCoreApplication>
#include <QDBusInterface>
#include <QDBusReply>


#define DBUS_TIMEOUT_MS 7500
/*******************************************************************************
    QDBusArgument operator
*******************************************************************************/
QDBusArgument& operator<< (QDBusArgument& arg, const QImage& image) 
{
  if (image.isNull()) {
    // Sometimes this gets called with a null QImage for no obvious reason.
    arg.beginStructure();
    arg << 0 << 0 << 0 << false << 0 << 0 << QByteArray();
    arg.endStructure();
    return arg;
  }
  QImage scaled = image.scaledToHeight(100, Qt::SmoothTransformation);
  QImage i = scaled.convertToFormat(QImage::Format_ARGB32).rgbSwapped();
  arg.beginStructure();
  arg << i.width();
  arg << i.height();
  arg << i.bytesPerLine();
  arg << i.hasAlphaChannel();
  int channels = i.isGrayscale() ? 1 : (i.hasAlphaChannel() ? 4 : 3);
  arg << i.depth() / channels;
  arg << channels;
  arg << QByteArray(reinterpret_cast<const char*>(i.bits()), i.numBytes());
  arg.endStructure();
  return arg;
}

const QDBusArgument& operator>> (const QDBusArgument& arg, QImage& image) {
  // This is needed to link but shouldn't be called.
  Q_ASSERT(0);
  return arg;
}

/*
********************************************************************************
*                                                                              *
*    Class DbusNotification                                                    *
*                                                                              *
********************************************************************************
*/
DbusNotification::DbusNotification(QObject* parent) : QObject(parent)
{
    init();
    reloadSettings();
}

void DbusNotification::init()
{
    _notification_id = 0;
    _interface       = 0;
    _t               = 0;  
    oldState         = ENGINE::STOPPED;
  
}

void DbusNotification::reloadSettings() 
{
     Debug::debug() << "DbusNotification -> reloadSettings";

    _enable       = SETTINGS()->_useDbusNotification;
    
    if(!_enable) {
      if(_interface) {
        disconnect(EnginePlayer::instance(), 0,this, 0);
        disconnect(MainWindow::instance(), 0,this, 0);  
        delete _interface;
      }
      if(_t) {
       disconnect(_t, 0,this, 0);
       delete _t;
      }
      init();
      return;
    }
    
    if(!_interface)
      _interface = new QDBusInterface("org.freedesktop.Notifications", "/org/freedesktop/Notifications");
    
    if(!_t)
      _t = new QTimer(this);
    
    connect(EnginePlayer::instance(), SIGNAL(engineStateChanged()), this, SLOT(handleStateChanged()));  
    connect(EnginePlayer::instance(), SIGNAL(volumeChanged(int)), this, SLOT(handleVolumeChanged()));
    connect(EnginePlayer::instance(), SIGNAL(trackChanged(const QString&)), this, SLOT(sendNowPlayingNotif()));
    connect(MainWindow::instance(), SIGNAL(playlistFinished()), this, SLOT(sendPlaylistFinished()));
    connect( _t, SIGNAL(timeout()), this, SLOT(sendVolumeChanged()) );
}


void DbusNotification::handleVolumeChanged()
{
    _t->stop();
    _t->start(3000);  
}

void DbusNotification::handleStateChanged() 
{
/* note 07-12-2012 : 
 * les messages pour la piste de lecture ne sont pas envoyés (mandriva + phonon-gstreamer)
 * il faut utiliser le signal  TrackChanged à la place du stateChange
 */
    ENGINE::E_ENGINE_STATE newState = EnginePlayer::instance()->engineState();
    if(oldState == newState)
      return;
  
    switch (newState) {
      case ENGINE::PLAYING :
        //! with phonon-gstreamer PLAYING state is not send for each trach 
        //! if playing is continuous (so use trach trackChanged signal)
        /*sendNowPlayingNotif(); */
        break;
      case ENGINE::PAUSED  : 
        //! phonon-vlc send pause state between track
        /*sendPaused();*/
        break;
      case ENGINE::STOPPED : sendStopped(); break;
      case ENGINE::ERROR   : sendStopped(); break;
    }        

    oldState = newState;
}

void DbusNotification::sendNowPlayingNotif()
{
    QString summary;
    QString message;
    QString icon = "notification-audio-play";
    
    //! -- Playing State
    MediaItemPtr mi = EnginePlayer::instance()->enginePlayingMediaItem();
    QPixmap     pix = EnginePlayer::instance()->enginePlayingPixmap();
    QImage    image = pix.toImage();
  
    if(mi->type == TYPE_TRACK) {
      //summary =  mi->data.at(TRACK::title).toString();
      summary =  mi->data.at(TRACK::title).toString();
      message = QString("%1 - %2").arg(
           mi->data.at(TRACK::artist).toString(),
           mi->data.at(TRACK::album).toString()
           );
    }
    else if(mi->type == TYPE_STREAM) {
      summary =  mi->data.at(STREAM::name).toString();     
      message = "";
    }
            
    //! send message
    sendMessage(summary,message,icon,image);      
}


void DbusNotification::sendPaused()     
{
    //! get now playing image
    QPixmap   pix = EnginePlayer::instance()->enginePlayingPixmap();
    if(pix.isNull()) 
      pix = QPixmap(":/icon/64x64/yarock.png");

    QImage  image = pix.toImage();
    QString icon = "notification-audio-play";

    //! send message
    sendMessage(QCoreApplication::applicationName(), tr("Paused"),icon,image);
}

void DbusNotification::sendStopped() 
{
    
    QPixmap pix = QPixmap(":/icon/64x64/yarock.png");
    QImage  image = pix.toImage();
    QString icon = "notification-audio-play";
    
    sendMessage(QCoreApplication::applicationName(), tr("Stopped"),icon,image);
}

//! TODO 
void DbusNotification::sendPlaylistFinished() 
{
    QPixmap pix = QPixmap(":/icon/64x64/yarock.png");
    QImage  image = pix.toImage();
    QString icon = "notification-audio-play";
    
    //! We get a PlaylistFinished followed by a Stopped from the player
    sendMessage(QCoreApplication::applicationName(), tr("Playlist finished"),icon,image);
}

void DbusNotification::sendVolumeChanged() 
{
    _t->stop();

    //! get now playing image
    QPixmap   pix = EnginePlayer::instance()->enginePlayingPixmap();
    if(pix.isNull()) 
      pix = QPixmap(":/icon/64x64/yarock.png");

    QImage  image = pix.toImage();
    QString icon = "notification-audio-play";
    const int value = EnginePlayer::instance()->volume();
    sendMessage(QCoreApplication::applicationName(), tr("Volume %1%").arg(value),icon,image);
}

//! ----- SendMessage ----------------------------------------------------------
void DbusNotification::sendMessage(
               const QString& summary, 
               const QString& message,
               const QString& icon, 
               const QImage& image) 
{
    //Debug::debug() << "DbusNotification -> sendMessage" << message;

    QVariantMap hints;
    if (!image.isNull())
      hints["image_data"] = QVariant(image);
    
    int id = 0;
    
    //! Reuse the existing popup if it's still open.  
    if (_last_notification_time.secsTo(QDateTime::currentDateTime()) * 1000 < DBUS_TIMEOUT_MS)      
      id = _notification_id;

    QString member = "Notify";
    QVariantList params;
    params << APP_NAME;
    params << static_cast<unsigned int>(id);
    params << icon;
    params << summary << message;
    params << QStringList();
    params << hints;
    params << DBUS_TIMEOUT_MS;
    
    QDBusPendingCall pcall = _interface->asyncCallWithArgumentList(member, params);
    QDBusPendingCallWatcher *watcher = new QDBusPendingCallWatcher(pcall, this);
    QObject::connect(watcher, SIGNAL(finished(QDBusPendingCallWatcher*)),
                  this, SLOT(callFinished(QDBusPendingCallWatcher*)));
}

//! ----- CallFinished ----------------------------------------------------------
void DbusNotification::callFinished(QDBusPendingCallWatcher* watcher)
{
    QDBusPendingReply<uint> reply = *watcher;
    if (reply.isError()) {
      qWarning() << "DbusNotification -> Error sending notification" << reply.error().name();
      return;
    }

    uint id = reply.value();
    if (id != 0) {
      _notification_id        = id;
      _last_notification_time = QDateTime::currentDateTime();
    }
}
