/* YaRock Player
* Copyright (C) 2010-2013 Sebastien amardeilh (sebastien.amardeilh+yarock@gmail.com)
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _MEDIAITEM_H_
#define _MEDIAITEM_H_

// Qt
#include <QVariant>
#include <QString>
#include <QSize>
#include <QList>
#include <QSharedData>
#include <QExplicitlySharedDataPointer>
#include <QMap>
#include <QVector> 

// taglib
#include <taglib/fileref.h>
#include <taglib/id3v2tag.h>
#include <taglib/attachedpictureframe.h>



//! MediaItem type
typedef enum {TYPE_UNKWON = 0, TYPE_ARTIST, TYPE_ALBUM, TYPE_TRACK, TYPE_TRACK_CUE, TYPE_STREAM, TYPE_PLAYLIST} T_TYPE;

//! MediaItem Playlist type
typedef enum {T_DATABASE = 0, T_FILE, T_SMART} T_PLAYLIST;


class MediaItem;
typedef QExplicitlySharedDataPointer<MediaItem> MediaItemPtr;
Q_DECLARE_METATYPE( MediaItemPtr )


/*
********************************************************************************
*                                                                              *
*    Class MediaItem                                                           *
*                                                                              *
********************************************************************************
*/
class MediaItem : public QSharedData 
{
  public:
    MediaItem(T_TYPE t = TYPE_TRACK);
    MediaItem(const MediaItem& other);
    ~MediaItem();
  
    T_TYPE                 type;
    QVector<QVariant>      data;
    QList<MediaItemPtr>    childItems;
    MediaItemPtr           parentItem;  
    
    // Simple accessors
    MediaItemPtr parent() { return parentItem;}
    void setParent(const MediaItemPtr p) { parentItem = p;}
    MediaItemPtr child(int index);
    int childCount() const;
    int childNumber() const;
    
    // Methods    
    bool removeChildren(int idx);
    MediaItemPtr addChildren(T_TYPE type);
  
  private:
    int sizeForType() const;
};

Q_DECLARE_METATYPE( MediaItem )



/*
********************************************************************************
*                                                                              *
*    namespace MEDIA                                                           *
*                                                                              *
********************************************************************************
*/
namespace MEDIA {
  void qResetAll(const QList<MediaItemPtr> mediaList);

  QString urlForMediaItem(const MediaItemPtr mi); 

  //! ------ Media Filter ------------------------------------------------------ 
  bool isAudioFile(const QString& url);
  bool isPlaylistFile(const QString& url);
  bool isCueFile(const QString& url);  
  bool isShoutCastUrl(const QString& url);

  //! ------ Remote Url management --------------------------------------------- 
  bool isUrlLocal(const QString& url);
  bool isMediaItemLocal(const MediaItemPtr mi); 

  //! ------ Read Tag with Taglib methods -------------------------------------- 
  QPixmap getPixmapFromMediaUrl(const QString& url, QSize size = QSize(128,128));
  QImage  getImageFromMediaUrl(const QString& url, QSize size);
    
  TagLib::ID3v2::AttachedPictureFrame *
  attachedPictureFrame(TagLib::ID3v2::Tag *id3Tag, bool create = false);

  // add detection of disk number FIXME (not very clean way)
  MediaItemPtr FromLocalFile(const QString url, int* p_disc=0);
  MediaItemPtr FromDataBase(const QString connectionName, const QString url);
  MediaItemPtr FromDataBase(const QString connectionName, int trackId);

  //! ------ CoverName Management ----------------------------------------------
  QString getCoverName(const MediaItemPtr mi);
  QString getCoverName(const QString& artist, const QString& album);
  
  //! ------ Media List management --------------------------------------------- 
  QString durationTextFromInt(const int MediaItemDuration) ;
  QString durationTextFromMediaList(const QList<MediaItemPtr> mediaList);
    
  //! --- sort utilities -------------------------------------------------------
  bool compareTrackItemGenre(const MediaItemPtr, const MediaItemPtr);
  bool compareAlbumItemYear(const MediaItemPtr, const MediaItemPtr);  
  bool compareAlbumItemPlaycount(const MediaItemPtr, const MediaItemPtr);
  bool compareAlbumItemRating(const MediaItemPtr, const MediaItemPtr);
  bool compareArtistItemPlaycount(const MediaItemPtr, const MediaItemPtr);
  bool compareArtistItemRating(const MediaItemPtr, const MediaItemPtr);
  bool compareStreamName(const MediaItemPtr, const MediaItemPtr);
  bool compareStreamCategorie(const MediaItemPtr, const MediaItemPtr);
  
  //! --- replay gain utilities ------------------------------------------------
  enum ReplayGainTag
  {
        ReplayGain_Track_Gain,
        ReplayGain_Track_Peak,
        ReplayGain_Album_Gain,
        ReplayGain_Album_Peak
  };  
  typedef QMap<ReplayGainTag, qreal> ReplayGainTagMap;
  ReplayGainTagMap readReplayGainTags( TagLib::FileRef fileref );

    //! --- statistic tags utilities ---------------------------------------------
  enum StatisticTag
  {
        Rating_Tag,
        Playcount_Tag
  };  
  typedef QMap<StatisticTag, int> StatisticTagMap;
  StatisticTagMap readStatisticTags( TagLib::FileRef fileref );
  
} // end namespace MEDIA
  

namespace ARTIST {
  enum E_FIELD { id = 0,          //! (int) database id 
                 name,            //! (QString) artist name
                 coverpathlist,   //! (QStringList) liste of path to album cover
                 isFavorite,      //! (Bool) favorite flag 
                 isPlaying,       //! (Bool) is playing flag
                 playcount,       //! (int) playcount integer
                 rating,          //! (float) artist rating
                 isUserRating     //! (bool) user or auto rating                 
                 };
}

namespace ALBUM {
  enum E_FIELD { id = 0,          //! (int) database id 
                 name,            //! (QString) album name
                 year,            //! (QString) album year
                 coverpath,       //! (QString) path to cover
                 isFavorite,      //! (Bool) favorite flag 
                 isPlaying,       //! (Bool) is playing flag
                 playcount,       //! (int) playcount integer         
                 rating,          //! (float) album rating
                 isUserRating     //! (bool) user or auto rating                 
                 };
}

namespace TRACK {
  enum E_FIELD { id = 0,          //! (int) database id 
                 url,             //! (QString) file path
                 name,            //! (QString) track short name
                 title,           //! (QString) track title
                 artist,          //! (QString) track artist name
                 album,           //! (QString) track album name
                 genre,           //! (QString) track genre name
                 duration,        //! (int) dure (second)
                 num,             //! (uint) Numero
                 year,            //! (uint) year
                 isPlaying,       //! (Bool) is playing flag
                 isBroken,        //! (Bool) is broken flag
                 lastPlayed,      //! (QString) last playing date from History 
                 isPlayed,        //! (Bool) is played flag
                 isStopAfter,     //! (Bool) flag to stop after                  
                 playcount,       //! (int) playcount integer
                 trackGain,       //! (qreal) track gain
                 trackPeak,       //! (qreal) track peak
                 albumGain,       //! (qreal) album gain
                 albumPeak,       //! (qreal) album peak
                 rating           //! (float) track rating
                 };
}

namespace TRACK_CUE {
  enum E_FIELD { id = 0,          //! (int) database id 
                 url,             //! (QString) file path
                 name,            //! (QString) track short name
                 title,           //! (QString) track title
                 artist,          //! (QString) track artist name
                 album,           //! (QString) track album name
                 genre,           //! (QString) track genre name
                 duration,        //! (int) dure (second)
                 num,             //! (uint) Numero
                 year,            //! (uint) year
                 isPlaying,       //! (Bool) is playing flag
                 isBroken,        //! (Bool) is broken flag
                 lastPlayed,      //! (QString) last playing date from History 
                 isPlayed,        //! (Bool) is played flag
                 isStopAfter,     //! (Bool) flag to stop after 
                 playcount,       //! (int) playcount integer
                 trackGain,       //! (qreal) track gain
                 trackPeak,       //! (qreal) track peak
                 albumGain,       //! (qreal) album gain
                 albumPeak,       //! (qreal) album peak
                 rating,          //! (float) track rating                                  
                 begin,           //! (int) 
                 end              //! (int) 
                 };
}


/* attention les id STREAM et TRACK correspondent */
namespace STREAM {
  enum E_FIELD { id = 0,          //! (int) database id 
                 url,             //! (QString) stream Url 
                 name,            //! (QString) stream short name
                 title,           //! (QString) stream current playing title
                 artist,          //! (QString) stream current playing artist
                 album,           //! (QString) stream current playing album
                 categorie,       //! (QString) stream categorie
                 isFavorite,      //! (Bool) is favorite flag
                 spare1,
                 spare2,
                 isPlaying,       //! (Bool) is playing flag
                 isBroken,        //! (Bool) is broken flag
                 lastPlayed,      //! (QString) last playing date from History 
                 isPlayed,        //! (Bool) is played flag
                 isStopAfter      //! (Bool) flag to stop after 
                 };
}

namespace PLAYLIST {
  enum E_FIELD { id = 0,          //! (int) database id 
                 url, 
                 name, 
                 icon,
                 rules,
                 date,
                 type,            //! (int) 0 = internal, 1 = user
                 isFavorite, 
                 isPlaying, 
                 isBroken
                 };
}

#endif //_MEDIAITEM_H_
